/*

*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <sys/time.h>

#include <netinet/in.h>

#include <limits.h>
#include <netdb.h>
#include <arpa/inet.h>

#include <sys/socket.h>
#include <errno.h>

#include <pthread.h>
#include <termios.h>

/* Text buffer sizes */

#define LINE_LEN	256
#define BUFFER_LEN	128

/* ----- Configurable items ----- */

unsigned short PORT;
char HOSTNAME[LINE_LEN];
int RATE;
char RFIDDEV[LINE_LEN];
int DELAY;
int VERBOSE;

/* =========== Subroutines ============= */


/* ------- Configuration routines  ------- */

void set_defaults()
{
    strcpy(HOSTNAME, "localhost");
    PORT = 1522;
    strcpy(RFIDDEV, "/dev/ttyS0");
    RATE = 2;
    DELAY = 0;
    VERBOSE = 0;
}


/* ---- Network routines ------- */

/* Converts ASCII text to in_addr struct */

struct in_addr *atoaddr(address)
char *address;
{
    struct hostent *host;
    static struct in_addr saddr;

    /* First try it as aaa.bbb.ccc.ddd. */
    saddr.s_addr = inet_addr(address);
    if (saddr.s_addr != -1) {
	return &saddr;
    }

    host = gethostbyname(address);
    if (host != NULL) {
	return (struct in_addr *) *host->h_addr_list;
    }
    return NULL;
}

/* Write buffer to socket connection */

int sock_write(int sockfd, char *buf, size_t count)
{
    size_t bytes_sent = 0;
    int this_write;

    while (bytes_sent < count) {
	do {
	    this_write = write(sockfd, buf, count - bytes_sent);
	} while ((this_write < 0) && (errno == EINTR));
	if (this_write <= 0) {
	    return (this_write);
	}
	bytes_sent += this_write;
	buf += this_write;
    }
    return (count);
}

/* Read from socket connection into buffer */

int sock_read(int sockfd, char *buf, size_t count)
{
    size_t bytes_read = 0;
    int this_read;
    fd_set readers;
    struct timeval timeout;
    int pending;

    while (bytes_read < count) {
	do {
	    /* Check status of connection */
	    pending = 0;
	    while (pending == 0) {
		FD_ZERO(&readers);
		FD_SET(sockfd, &readers);
		timeout.tv_sec = 5;
		timeout.tv_usec = 0;
		pending = select(FD_SETSIZE, &readers, 0, 0, &timeout);
		if (pending < 0) {
		    fprintf(stderr, "Connection closed ...\n");
		    return (-1);
		}
	    }
	    this_read = read(sockfd, buf, count - bytes_read);
	} while ((this_read < 0) && (errno == EINTR));
	if ((this_read == 0) && (pending == 1)) {
	    fprintf(stderr, "Connection closed ...\n");
	    return (-1);
	}
	if (this_read <= 0) {
	    return (this_read);
	}
	bytes_read += this_read;
	buf += this_read;
    }
    return (count);
}

/* Read a string delimited by a newline character from a socket connection */

int sock_gets(int sockfd, char *str, size_t count)
{
    int bytes_read;
    int total_count = 0;
    char *current_position;
    char last_read = 0;
    fd_set readers;
    struct timeval timeout;
    int pending;

    current_position = str;
    while ((last_read != 10) && (last_read != 13)) {
	/* Check status of connection */
	pending = 0;
	while (pending == 0) {
	    FD_ZERO(&readers);
	    FD_SET(sockfd, &readers);
	    timeout.tv_sec = 5;
	    timeout.tv_usec = 0;
	    pending = select(FD_SETSIZE, &readers, 0, 0, &timeout);
	    if (pending < 0) {
		fprintf(stderr, "Connection closed ...\n");
		return (-1);
	    }
	}
	bytes_read = read(sockfd, &last_read, 1);
	if ((bytes_read == 0) && (pending == 1)) {
	    fprintf(stderr, "Connection closed ...\n");
	    return (-1);
	}
	if (bytes_read <= 0) {
	    return bytes_read;
	}
	if ((total_count < count) && (last_read != 10) && (last_read != 13)) {
	    current_position[0] = last_read;
	    current_position++;
	    total_count++;
	}
    }
    if (count > 0)
	current_position[0] = 0;
    return total_count;
}

/* ---------- RFID routines --------- */

/* Send RFID command */

int rfidSend(int rfid_fd)
{
    unsigned char message[5];
#ifdef DEBUG
    int i;
#endif

    /* Setup command message */
    message[0] = 0x01;
    message[1] = 0x02;
    message[2] = 0x08;		// 0x09 normal mode
    message[3] = 0x32;		// power burst length (ms)
    message[4] = 0x38;

    if (write(rfid_fd, message, 5) != 5) {
	perror("write()");
	return (-1);
    } else {

#ifdef DEBUG
	fprintf(stderr, "Wrote:");
	for (i = 0; i < 5; i++) {
	    fprintf(stderr, "%02x ", (unsigned char) message[i]);
	}
	fprintf(stderr, "\n");
#endif

        if (DELAY) {
	 usleep(DELAY);
        }
        
	return (0);
    }
}

/* Receive RFID command */

int rfidReceive(int rfid_fd, unsigned int *dataA, unsigned int *dataB)
{
    unsigned char abyte;
    unsigned char message[256];
    unsigned char bcc;
    int i;
    int num_read, total_read, to_read;

    // Reset message data
    memset(message, 0, 12);
    total_read = 0;
    *dataA = 0;
    *dataB = 0;

    // Look for start mark
    while (abyte != 0x01) {
	num_read = read(rfid_fd, &abyte, 1);
	if (num_read < 0) {
	    perror("read()");
	    return (-1);
	}
#ifdef DEBUG
	if (abyte != 0x01) {
	    fprintf(stderr, "Read garbage byte: %02x\n", abyte);
	}
#endif
	if (DELAY) {
	    usleep(DELAY);
	}
    }

    // Read number of bytes to read
    num_read = read(rfid_fd, &abyte, 1);
    if (num_read < 0) {
	perror("read()");
	return (-1);
    }
    if (DELAY) {
	usleep(DELAY);
    }
    to_read = abyte;

    // Check number of bytes to read
#ifdef DEBUG
    if ((to_read != 1) && (to_read != 9)) {
	fprintf(stderr, "Bad message length: %i bytes\n", to_read);
    }
#endif

    // Read data
    while (total_read < to_read) {
	num_read = read(rfid_fd, &message[total_read], 1);
	if (num_read < 0) {
	    perror("read()");
	    return (-1);
	}
	total_read++;
	if (DELAY) {
	    usleep(DELAY);
	}
    }

#ifdef DEBUG
    fprintf(stderr, "Read %i bytes: ", total_read);
    for (i = 0; i < total_read; i++) {
	fprintf(stderr, "%02x ", message[i]);
    }
    fprintf(stderr, "\n");
#endif


    // Read BCC
    num_read = read(rfid_fd, &abyte, 1);
    if (num_read < 0) {
	perror("read()");
	return (-1);
    }
    if (DELAY) {
	usleep(DELAY);
    }

    // Calculate BCC
    bcc=to_read;
    for (i=0; i<total_read; i++) {
     bcc = bcc ^ message[i];
    }

#ifdef DEBUG
    fprintf (stderr,"BCC: Send=%i Calculated=%i\n",abyte,bcc);
#endif
    
    if ((total_read == 9) && ((message[0] == 0x0c) || (message[0] == 0x0d)) && (abyte==bcc)) {
	/* Copy data byte */
	*dataB = (unsigned int) (message[1] + 256 * message[2] + (256 * 256) * message[3] + (256 * 256 * 256) * message[4]);
	*dataA = (unsigned int) (message[5] + 256 * message[6] + (256 * 256) * message[7] + (256 * 256 * 256) * message[8]);

#ifdef DEBUG
	fprintf(stderr, "Calculated: A=%u B=%u\n", *dataA, *dataB);
#endif
    }

    return (0);
}

/* ======== Main processing loop ============= */

void NetworkLoop(int connected_socket, int rfid_fd)
{
    int counter = 0;
    int connected;
    char outbuffer[BUFFER_LEN];
    int USLEEP;
    unsigned int dataA, dataB;

    /* Calculate rate-throttle pause */
    USLEEP = 1000000 / RATE;
    /* Processing loop */
    connected = 1;
    while (connected) {
	/* Count iterations */
	counter++;

	/* Read RFID */
	rfidSend(rfid_fd);
	if (rfidReceive(rfid_fd, &dataA, &dataB) == 0) {

	    /* Create output line */
	    if ( (dataA != 0) || (dataB != 0) ) {
		sprintf(outbuffer, "PUSHrfid=%08X%08X\n", dataA,dataB);

                if (VERBOSE) {
 		 fprintf(stderr, "%08X%08X\n", dataA,dataB);
                }

#ifdef DEBUG
		fprintf(stderr, "Reflector Message: %s", outbuffer);
#endif
		/* Write POLL to reflector */
		if (connected_socket) {
		    connected = (sock_write(connected_socket, outbuffer, strlen(outbuffer)) > 0);
		}
            }
	}

	/* Pause */
	usleep(USLEEP);
    }
}

/* Main routine */

int main(int argc, char *argv[])
{
    int i;
    struct sockaddr_in address;
    static struct in_addr *addr;
    int connected;
    int connected_socket = -1;
    int rfid_fd;
    struct termios newtio, oldtio;

    /* Print message */
    printf("rfid2reflector - A. Schiffler, 03/2005\n\n");

    /* Initialize variables */
    set_defaults();

    /* Parse command line */
    for (i = 1; i < argc; i++) {
	char *arg;
	if (argv[i][0] != '-')
	    break;
	arg = argv[i];
	arg++;
	if (*arg == '-')
	    arg++;
	if (strncmp(arg, "port", 4) == 0) {
	    i++;
	    if (i <= argc) {
		PORT = atoi(argv[i]);
	    }
	} else if (strncmp(arg, "verbose", 7) == 0) {
            VERBOSE=1;
	} else if (strncmp(arg, "rate", 4) == 0) {
	    i++;
	    if (i <= argc) {
		RATE = atoi(argv[i]);
	    }
	} else if (strncmp(arg, "delay", 5) == 0) {
	    i++;
	    if (i <= argc) {
		DELAY = atoi(argv[i]);
	    }
	} else if (strncmp(arg, "rfiddev", 7) == 0) {
	    i++;
	    if (i <= argc) {
		strcpy(RFIDDEV, argv[i]);
	    }
	} else if (strncmp(arg, "host", 4) == 0) {
	    i++;
	    if (i <= argc) {
		strcpy(HOSTNAME, argv[i]);
	    }
	} else {
	    /* Command line help */
	    printf("Command line parameters:\n");
	    printf("\nNetwork settings:\n");
	    printf(" -host name      host to connect to (default: %s)\n", HOSTNAME);
	    printf(" -port #         port to connect to (default: %i)\n", PORT);
	    printf("\nOutput settings:\n");
	    printf(" -rfiddev dev    RFID device for scanning (default: %s)\n", RFIDDEV);
	    printf(" -rate #         update rate for scanning device in Hz (default: %i)\n", RATE);
	    printf(" -delay #        RS232 inter-char read delay (usec) (default: %i)\n", DELAY);
	    printf(" -verbose        show RFIDs read (default: off)\n");	    
	    exit(0);
	}
    }

    /* Setup internet address information */
    addr = (struct in_addr *) atoaddr((char *) HOSTNAME);
    if (addr == NULL) {
	fprintf(stderr, "Invalid network address: %s", HOSTNAME);
	exit(-1);
    }
    memset((char *) &address, 0, sizeof(address));
    address.sin_family = AF_INET;
    address.sin_addr.s_addr = addr->s_addr;
    address.sin_port = htons(PORT);

    /* Connect to reflector */
    connected_socket = socket(AF_INET, SOCK_STREAM, 0);
    printf("Connecting to %s on port %d ...", inet_ntoa(*addr), PORT);
    connected = connect(connected_socket, (struct sockaddr *) &address, sizeof(address));
    if (connected < 0) {
	printf("\n");
	perror("connect()");
	connected_socket = 0;
    }
    printf(" connected.\n");

    /* Open RFID port */
    printf("Opening RFID device %s.\n", RFIDDEV);
    if ((rfid_fd = open(RFIDDEV, O_RDWR | O_NOCTTY)) < 0) {
	perror("open()");
	if (connected_socket > 0) {
	    close(connected_socket);
	}
	exit(-1);
    }

    /* Termial setup */
    tcgetattr(rfid_fd, &oldtio);
    bzero(&newtio, sizeof(newtio));
    newtio.c_cflag = B9600 | CRTSCTS | CS8 | CLOCAL | CREAD;
    newtio.c_iflag = IGNPAR;
    newtio.c_oflag = 0;
    newtio.c_lflag = 0;
    newtio.c_cc[VTIME] = 0;
    newtio.c_cc[VMIN] = 1;
    tcflush(rfid_fd, TCIFLUSH);
    tcsetattr(rfid_fd, TCSANOW, &newtio);

    /* Start processing the reflector text and generating MIDI controls */
    printf("Processing ...\n");
    NetworkLoop(connected_socket, rfid_fd);

    /* Close file/socket */
    if (connected_socket > 0) {
	close(connected_socket);
    }
    close(rfid_fd);


    exit(0);
}
